#!/usr/bin/env python

"""
This script will emulate a Sprint executable, so that we can use it for SprintDatasetBase.
This is useful for tests.
To generate data, we can use the GeneratingDataset code.
"""

from __future__ import annotations

import sys
from importlib import import_module

import _setup_test_env  # noqa
import returnn.datasets.generating as generating_dataset
from returnn.datasets import Dataset
from returnn.util.basic import ObjAsDict


class ArgParser:
    """
    Emulate the Sprint argument parser.
    """

    def __init__(self):
        self.args = {}

    def add(self, key, value):
        """
        :param str key:
        :param str value:
        """
        self.args[key] = value

    def get(self, key, default=None):
        """
        :param str key:
        :param str|T|None default:
        :rtype: str|T|None
        """
        return self.args.get(key, default)

    def parse(self, argv):
        """
        :type argv: list[str]
        """
        i = 0
        while i < len(argv):
            arg = argv[i]
            if arg.startswith("--"):
                key, value = arg[2:].split("=", 1)
                if key.startswith("*."):  # simple hack. example "--*.input-dim=5"
                    key = key[2:]
                self.add(key, value)
            i += 1


def main(argv):
    """
    Main entry.
    """
    print("DummySprintExec init", argv)
    args = ArgParser()
    args.parse(argv[1:])

    if args.get("pymod-name"):
        sprint_api = import_module(args.get("pymod-name"))
    else:
        import returnn.sprint.extern_interface as sprint_api

    input_dim = int(args.get("feature-dimension"))
    assert input_dim > 0
    output_dim = int(args.get("trainer-output-dimension"))
    assert output_dim > 0
    sprint_config = args.get("pymod-config", "")
    target_mode = args.get("target-mode", "target-generic")
    sprint_api.init(inputDim=input_dim, outputDim=output_dim, config=sprint_config, targetMode=target_mode)

    if args.get("crnn-dataset"):
        dataset = eval(args.get("crnn-dataset"), {}, ObjAsDict(generating_dataset))
        assert isinstance(dataset, Dataset)
        assert dataset.num_inputs == input_dim
        assert dataset.num_outputs == {"classes": (output_dim, 1), "data": (input_dim, 2)}
        dataset.init_seq_order(epoch=1)

        seq_idx = 0
        while dataset.is_less_than_num_seqs(seq_idx):
            dataset.load_seqs(seq_idx, seq_idx + 1)
            features = dataset.get_data(seq_idx, "data")
            features = features.T  # Sprint-like
            kwargs = {"features": features}
            if target_mode == "target-generic":
                if "orth" in dataset.get_target_list():
                    kwargs["orthography"] = dataset.get_data(seq_idx, "orth")
                if "classes" in dataset.get_target_list():
                    kwargs["alignment"] = dataset.get_data(seq_idx, "classes")
                print("DummySprintExec seq_idx %i feedInputAndTarget(**%r)" % (seq_idx, kwargs))
                sprint_api.feedInputAndTarget(**kwargs)
            else:
                raise NotImplementedError("targetMode = %s" % target_mode)
            seq_idx += 1

    print("DummySprintExec exit")
    sprint_api.exit()


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    main(sys.argv)



"""
Utilities which affect the gradient
"""

from __future__ import annotations
from typing import Optional, Union
from returnn.tensor import Tensor, Dim


__all__ = ["set_requires_gradient", "gradient", "stop_gradient", "scaled_gradient", "scaled_gradient_ext"]


def set_requires_gradient(source: Tensor):
    """
    :param source:
    :return: nothing, modifies source in-place
    """
    # noinspection PyProtectedMember
    return source._raw_backend.set_requires_gradient(source)


def gradient(y: Tensor, x: Tensor) -> Tensor:
    """
    :param y: some scalar
    :param x: some tensor
    :return: gradient of y w.r.t. x
    """
    # noinspection PyProtectedMember
    return y._raw_backend.gradient(y, x)


def stop_gradient(source: Tensor) -> Tensor:
    """wraps tf.stop_gradient or torch detach"""
    # noinspection PyProtectedMember
    return source._raw_backend.stop_gradient(source)


def scaled_gradient(source: Tensor, scale: Union[float, Tensor]) -> Tensor:
    """
    :param source:
    :param scale: if constant 0., will use :func:`stop_gradient`.
        Can be used as gradient reversal layer (with negative factor).
    :return: source with scaled gradient
    """
    if not isinstance(scale, Tensor) and scale == 0.0:
        return stop_gradient(source)
    # noinspection PyProtectedMember
    return source._raw_backend.scaled_gradient(source, scale)


def scaled_gradient_ext(
    source: Tensor,
    *,
    scale: Union[float, Tensor],
    shift: Optional[Union[float, Tensor]] = None,
    scale_shift_by_sum_over_axis: Optional[Dim] = None,
) -> Tensor:
    """
    Just `identity` in the forward pass.
    Scales the gradient by some factor in backprop.
    Can be used as gradient reversal layer (with negative factor).
    For TF, uses :func:`returnn.tf.util.basic.scaled_gradient`, or :func:`tf.stop_gradient`

    :param source:
    :param scale: if constant 0. and no shift, will use :func:`stop_gradient`
    :param shift:
    :param scale_shift_by_sum_over_axis: if given, calculates the sum over this axis (absolute values)
        and multiplies the shift value by this sum.
    :return: source with transformed gradient
    """
    # noinspection PyProtectedMember
    return source._raw_backend.scaled_gradient_ext(
        source, scale=scale, shift=shift, scale_shift_by_sum_over_axis=scale_shift_by_sum_over_axis
    )



"""
Provides :class:`Data`, :class:`Dim`, :class:`SearchBeam`.

See :ref:`data` for some higher-level description.
"""

from __future__ import annotations

from typing import Optional, Union, Dict, List, Tuple
import typing
import os
import tensorflow as tf

from returnn.util.basic import NotSpecified

# Import also batch_dim, single_step_dim, to support old code.
# noinspection PyUnresolvedReferences
from returnn.tensor import Tensor, Dim, batch_dim, single_step_dim, ControlFlowContext

# Import also more to support old code.
# noinspection PyUnresolvedReferences
from returnn.tensor.marked_dim import MarkedDim, ImplicitSparseDim, ImplicitDynSizeDim, OptionalDim

# Import to support old code.
# noinspection PyUnresolvedReferences
from returnn.tensor.dim import VerifyOutShapeException


# Alias for old code compatibility.
Data = Tensor

# Earlier the class was called DimensionTag. Provide this alias for older code.
DimensionTag = Dim

# Alias for older code.
_MarkedDim = MarkedDim


# Provide some simple wrappers. https://github.com/rwth-i6/returnn/issues/782
# Use CamelCase function names (invalidates PEP8) to make it look like a class instance.

# noinspection PyPep8Naming
def FeatureDim(description, dimension, **kwargs):
    """
    DEPRECATED. Use :class:`Dim` instead, and setting the `kind` is not needed anymore.

    :param str description:
    :param int|None dimension:
    :rtype: Dim
    """
    return Dim(kind=Dim.Types.Feature, description=description, dimension=dimension, **kwargs)


# noinspection PyPep8Naming
def SpatialDim(description, dimension=None, **kwargs):
    """
    DEPRECATED. Use :class:`Dim` instead, and setting the `kind` is not needed anymore.

    :param str description:
    :param int|None dimension:
    :rtype: Dim
    """
    return Dim(kind=Dim.Types.Spatial, description=description, dimension=dimension, **kwargs)


class BatchInfo:
    """
    A batched tensor is a tensor with batch dimension,
    i.e. consisting of multiple samples/sequences
    which are supposed to be totally independent of each other.

    The batch dim can consists out of one or more flattened "virtual" dims,
    which :class:`BatchInfo` keeps track of.
    This class provides some additional information
    about the batch dimension.
    Only one instance per different type of batch-dim is expected
    (i.e. `batch_info1 is batch_info2` <==> same batch info).

    When we pass data from the dataset to the network
    (in all cases (training, inference ...) via :class:`Runner` in the TF engine),
    we get a batch dimension due to the minibatch construction.
    This is a global batch size
    (usually dynamic, because every minibatch/step can have a different amount of samples,
    although we can also support static sizes, which is needed e.g. for TPUs)
    represented by :class:`BatchInfo.GlobalBatchDim`.

    When we do beam search (see :class:`SearchBeam`),
    we have multiple hypotheses per batch item,
    and thus a different batch dimension.

    We can also pack arrays (multiple sequences)
    (also referred to as "flattened tensors", "non-padded tensors", "ragged tensors").
    See e.g. :class:`FlattenBatchLayer` or :func:`flatten_with_seq_len_mask`.
    Also see :class:`tf.RaggedTensor` which also represents
    packed tensors (but only batch-major, although this is just a reinterpretation).
    We do not directly use :class:`tf.RaggedTensor` in :class:`Data`
    to have robust and reliable code (which expects :class:`tf.Tensor`).
    However, we maybe can make use of some of the functions in :mod:`tf.ragged`.
    """

    class VirtualDimBase(object):
        """
        Represents one virtual dim, flattened into the batch dim.
        """

        def short_repr(self):
            """
            :rtype: str
            """
            raise NotImplementedError

        def __repr__(self):
            return "%s{%s}" % (self.__class__.__name__, self.short_repr())

    class FixedDim(VirtualDimBase):
        """
        Represents a dim with fixed size.
        """

        def __init__(self, size: Union[tf.Tensor, int], dim_tag: Optional[Dim] = None):
            """
            :param size:
            :param dim_tag:
            """
            self.size = size
            self.dim_tag = dim_tag

        def short_repr(self):
            """
            :rtype: str
            """
            if isinstance(self.size, int):
                return "F(%i)" % self.size
            return "F(?)"

    class GlobalBatchDim(FixedDim):
        """
        Represents the global batch dim by the network (minibatch construction from the dataset).
        """

        def __init__(self, size: Union[tf.Tensor, int], dim_tag: Optional[Dim] = None):
            if not dim_tag:
                dim_tag = batch_dim  # use global batch dim tag
            super().__init__(size=size, dim_tag=dim_tag)

        def short_repr(self):
            """
            :rtype: str
            """
            if isinstance(self.size, int) and self.size >= 0:
                return "B(%i)" % self.size
            return "B"

    class BeamDim(FixedDim):
        """
        Represents a search beam.
        """

        def __init__(self, beam):
            """
            :param SearchBeam beam:
            """
            super(BatchInfo.BeamDim, self).__init__(size=beam.beam_size)
            self.beam = beam

        def short_repr(self):
            """
            :rtype: str
            """
            return "Beam{%r}(%s)" % (self.beam.name, self.size)

    class PaddedDim(FixedDim):
        """
        Represents a dim with variable size, which is flattened with padding (not packed) into the batch.
        """

        def __init__(self, dim_tag):
            """
            :param Dim dim_tag:
            """
            super(BatchInfo.PaddedDim, self).__init__(size=dim_tag.get_dim_value())
            self.dim_tag = dim_tag

        def short_repr(self):
            """
            :rtype: str
            """
            return "Padded{%r}" % self.dim_tag.description

    class PackedDim(VirtualDimBase):
        """
        Represents a dim with variable sizes, which is packed (un-padded) into the batch.
        Variable w.r.t. other dims (must be per batch entry).
        """

        def __init__(self, dim_tag, key_axes):
            """
            :param Dim dim_tag:
            :param list[BatchInfo.VirtualDimBase] key_axes:
              most common case would be [GlobalBatchDim(...)],
              but [GlobalBatchDim(...),BeamDim(...)] is also common.
            """
            self.dim_tag = dim_tag
            self.key_axes = key_axes

        @property
        def sizes(self):
            """
            :return: shape [B_flat]
            :rtype: tf.Tensor
            """
            assert self.dim_tag.dyn_size is not None
            return self.dim_tag.dyn_size

        def short_repr(self):
            """
            :rtype: str
            """
            return "Packed{%r}" % (self.dim_tag.description,)

    def __init__(self, base, new_dim, new_dim_index=None):
        """
        :param BatchInfo|None base:
          If this is extended or based on another batch.
          Except of the batch dim of the dataset minibatch,
          we would always have a base.
        :param BatchInfo.VirtualDimBase|None new_dim:
        :param int|None new_dim_index:

        In most cases, this constructor would probably not be used directly by the user.
        """
        self.base = base
        virtual_dims = list(base.virtual_dims) if base else []
        if new_dim:
            if new_dim_index is None:
                assert not virtual_dims
                new_dim_index = 0
            if new_dim_index < 0:
                assert new_dim_index == -1
                virtual_dims.append(new_dim)
            else:
                virtual_dims.insert(new_dim_index, new_dim)
        self.virtual_dims = virtual_dims  # type: typing.List[BatchInfo.VirtualDimBase]
        self._dim = None  # type: typing.Optional[typing.Union[tf.Tensor,int]]
        self.batch_dim_tag: Optional[Dim] = None
        if not base and isinstance(new_dim, BatchInfo.GlobalBatchDim):
            self.batch_dim_tag = new_dim.dim_tag
        else:
            self.batch_dim_tag = Dim(
                kind=Dim.Types.Batch, description="batch:%s" % self.short_repr(), batch=self, dimension=self.static_dim
            )
        # These self._global_... attributes are meant
        # to be accessed only via the global (root) object (via get_global_base).
        # They store global information.
        # We don't use class attributes because this should not be global per process but only per network.
        self._global_beam_dims_by_beam_name = {}  # type: Dict[str,BatchInfo.BeamDim]
        self._global_padded_dims_by_dim_tag = {}  # type: Dict[Dim,BatchInfo.PaddedDim]
        self._packed_dims_by_dim_tag = {}  # type: Dict[Dim,BatchInfo.PackedDim]
        self.descendants = []  # type: List[BatchInfo]
        self._descendants_by_beam_name = {}  # type: Dict[str,BatchInfo]
        self._global_descendants_by_virtual_dims = {}  # type: Dict[Tuple[BatchInfo.VirtualDimBase,...],BatchInfo]
        if base:
            base.descendants.append(self)
            if isinstance(new_dim, BatchInfo.BeamDim):
                beam = new_dim.beam
                assert beam.name not in base._descendants_by_beam_name
                base._descendants_by_beam_name[beam.name] = self
        global_base = self.get_global_base()
        assert tuple(self.virtual_dims) not in global_base._global_descendants_by_virtual_dims
        global_base._global_descendants_by_virtual_dims[tuple(self.virtual_dims)] = self

    # noinspection PyShadowingNames
    @classmethod
    def make_global_batch_info(cls, batch_dim):
        """
        :param tf.Tensor|int batch_dim:
        :return: global batch info w.r.t. the network / graph
        :rtype: BatchInfo
        """
        # This is not stored in a class attrib because this is only w.r.t. the network, not global in the process.
        return BatchInfo(base=None, new_dim=BatchInfo.GlobalBatchDim(size=batch_dim))

    _global_broadcast_batch = None  # type: typing.Optional[BatchInfo]

    @classmethod
    def make_global_broadcast_batch_info(cls):
        """
        :return: BatchInfo with no virtual dims, s.t. the dimension is 1 (== prod([])) (broadcastable)
        :rtype: BatchInfo
        """
        if cls._global_broadcast_batch:
            return cls._global_broadcast_batch
        cls._global_broadcast_batch = BatchInfo(base=None, new_dim=None)
        return cls._global_broadcast_batch

    @classmethod
    def get_common_batch_info(cls, batches):
        """
        :param list[BatchInfo|None] batches:
        :rtype: BatchInfo|None
        """
        # Fast paths.
        if not batches:
            return None
        if len(batches) == 1:
            return batches[0]
        # Make unique, and filter non-none.
        batches_ = []
        for batch in batches:
            if batch and batch not in batches_:
                batches_.append(batch)
        batches = batches_
        if not batches_:
            return None
        if len(batches) == 1:
            return batches[0]
        base = batches[0].get_global_base()

        # Collect all dims.
        all_virtual_dims = []
        for batch in batches:
            for dim in batch.virtual_dims:
                if dim not in all_virtual_dims:
                    # We want to get a reasonable order.
                    same_type_last_idx = None
                    for i, dim_ in enumerate(all_virtual_dims):
                        if type(dim_) == type(dim):
                            same_type_last_idx = i
                    if same_type_last_idx is not None:
                        all_virtual_dims.insert(same_type_last_idx + 1, dim)
                    else:
                        all_virtual_dims.append(dim)

        # Check if some batch already has them all.
        for batch in batches:
            if set(batch.virtual_dims) == set(all_virtual_dims):  # allow different order here
                return batch

        # Ok, need to extend.
        global_batch_dims = [dim for dim in all_virtual_dims if isinstance(dim, BatchInfo.GlobalBatchDim)]
        assert len(global_batch_dims) == 1
        global_batch_dim = global_batch_dims[0]
        assert base.virtual_dims == [global_batch_dim]
        beams = [dim for dim in all_virtual_dims if isinstance(dim, BatchInfo.BeamDim)]
        if beams:
            base = base.copy_extend_with_beam(SearchBeam.get_combined_beam(*(b.beam for b in beams)))
        dim_idx = 0
        for dim in all_virtual_dims:
            if dim in global_batch_dims:
                dim_idx += 1 + len(beams)
                continue
            if dim in beams:
                continue
            base = base._copy_extend_dim(new_dim=dim, new_dim_idx=dim_idx)
            dim_idx += 1
        return base

    def __repr__(self):
        return "BatchInfo{%s}" % ", ".join([dim.short_repr() for dim in self.virtual_dims])

    def short_repr(self):
        """
        :rtype: str
        """
        # "x" is the Theano-style shortcut for a broadcast dim.
        return "&".join([dim.short_repr() for dim in self.virtual_dims] or ["Bx"])

    def __getstate__(self):
        raise Exception("Pickling of BatchInfo is not supported. (%s)" % self)

    @property
    def dim(self):
        """
        :rtype: tf.Tensor|int
        """
        if self._dim is not None:
            return self._dim
        if not self.virtual_dims:
            return 1
        if len(self.virtual_dims) == 1:
            dim = self.virtual_dims[0]
            assert isinstance(dim, BatchInfo.FixedDim)
            return dim.size
        from returnn.tf.util.basic import same_control_flow_ctx, optional_mul

        if all(isinstance(dim, BatchInfo.FixedDim) for dim in self.virtual_dims):
            dims = self.virtual_dims  # type: typing.List[BatchInfo.FixedDim]
            sizes = [dim.size for dim in dims]
            with same_control_flow_ctx(sizes):
                value = optional_mul(*sizes)  # type: typing.Union[tf.Tensor,int]
            self._dim = value
            return value
        if all(isinstance(dim, (BatchInfo.PackedDim, BatchInfo.GlobalBatchDim)) for dim in self.virtual_dims):
            dims = [dim for dim in self.virtual_dims if isinstance(dim, BatchInfo.PackedDim)]
            if len(dims) > 1:
                raise NotImplementedError("%s: currently only support one packed dim but have %r" % (self, dims))
            (dim,) = dims
            assert isinstance(dim, BatchInfo.PackedDim)
            with same_control_flow_ctx(dim.dim_tag.dyn_size_ext.placeholder):
                value = tf.reduce_sum(dim.dim_tag.dyn_size_ext.placeholder)
            self._dim = value
            return value
        raise NotImplementedError("%r.dim()" % self)

    @dim.setter
    def dim(self, value):
        """
        Can only set the global batch dim.

        :param tf.Tensor|int value:
        """
        assert len(self.virtual_dims) == 1
        dim = self.virtual_dims[0]
        assert isinstance(dim, BatchInfo.GlobalBatchDim)
        dim.size = value
        if dim.dim_tag:
            dim.dim_tag.capacity = dim.dim_tag.size = value if (isinstance(value, int) and value > 0) else None
        self._dim = value

    @property
    def static_dim(self):
        """
        :rtype: int|None
        """
        # This should be safe. Do not call self.dim.
        if self._dim is not None:
            return self._dim if isinstance(self._dim, int) else None
        if not self.virtual_dims:
            return 1
        if len(self.virtual_dims) == 1:
            dim = self.virtual_dims[0]
            assert isinstance(dim, BatchInfo.FixedDim)
            return dim.size if isinstance(dim.size, int) else None
        from functools import reduce
        from operator import mul

        if all(isinstance(dim, BatchInfo.FixedDim) for dim in self.virtual_dims):
            dims = self.virtual_dims  # type: typing.List[BatchInfo.FixedDim]
            sizes = [dim.size for dim in dims]
            if all(isinstance(s, int) for s in sizes):
                return reduce(mul, sizes, 1)
            return None
        return None

    @property
    def beam(self):
        """
        :rtype: SearchBeam|None
        """
        beams = [dim for dim in self.virtual_dims if isinstance(dim, BatchInfo.BeamDim)]
        if beams:
            # Just return first. In case you need more custom logic, directly check the dims.
            return beams[0].beam
        return None

    def get_base_chain(self):
        """
        :rtype: list[BatchInfo]
        """
        bases = []
        base = self.base
        while base:
            bases.append(base)
            base = base.base
        return bases

    def get_global_base(self):
        """
        :rtype: BatchInfo
        """
        if not self.base:
            return self
        return self.get_base_chain()[-1]

    def get_global_batch_dim(self):
        """
        :rtype: BatchInfo.GlobalBatchDim
        """
        global_beam_dims = [dim for dim in self.virtual_dims if isinstance(dim, BatchInfo.GlobalBatchDim)]
        assert len(global_beam_dims) == 1
        return global_beam_dims[0]

    def is_global_batch(self):
        """
        :rtype: bool
        """
        global_beam_dims = [dim for dim in self.virtual_dims if isinstance(dim, BatchInfo.GlobalBatchDim)]
        return len(global_beam_dims) == 1 and len(self.virtual_dims) == 1

    def is_broadcast(self):
        """
        :rtype: bool
        """
        return len(self.virtual_dims) == 0

    def _make_beam_dim(self, beam):
        """
        :param SearchBeam beam:
        :rtype: BatchInfo.BeamDim
        """
        assert self.virtual_dims
        root = self.get_global_base()
        if beam.name in root._global_beam_dims_by_beam_name:
            return root._global_beam_dims_by_beam_name[beam.name]
        new_dim = BatchInfo.BeamDim(beam=beam)
        root._global_beam_dims_by_beam_name[beam.name] = new_dim
        return new_dim

    def _make_packed_dim(self, dim_tag):
        """
        :param Dim dim_tag:
        :rtype: BatchInfo.PackedDim
        """
        assert self.virtual_dims
        assert dim_tag.dyn_size is not None
        dim_tag_base = dim_tag.get_same_base()
        if dim_tag_base in self._packed_dims_by_dim_tag:
            return self._packed_dims_by_dim_tag[dim_tag_base]
        new_dim = BatchInfo.PackedDim(dim_tag=dim_tag, key_axes=self.virtual_dims)
        self._packed_dims_by_dim_tag[dim_tag_base] = new_dim
        return new_dim

    def _make_padded_dim(self, dim_tag):
        """
        :param Dim dim_tag:
        :rtype: BatchInfo.PaddedDim
        """
        assert self.virtual_dims
        root = self.get_global_base()
        assert dim_tag.dyn_size is not None
        dim_tag_base = dim_tag.get_for_batch_ctx(self, dim_tag.control_flow_ctx)
        if dim_tag_base in root._global_padded_dims_by_dim_tag:
            return root._global_padded_dims_by_dim_tag[dim_tag_base]
        new_dim = BatchInfo.PaddedDim(dim_tag=dim_tag_base)
        root._global_padded_dims_by_dim_tag[dim_tag_base] = new_dim
        return new_dim

    def _next_spatial_major_index(self):
        idx = None
        for i, dim in enumerate(self.virtual_dims):
            if isinstance(dim, BatchInfo.GlobalBatchDim):
                break
            if isinstance(dim, BatchInfo.BeamDim):
                break
            assert isinstance(dim, BatchInfo.FixedDim)
            idx = i + 1
        if idx is not None:
            return idx
        return 0

    def copy_extend_with_beam(self, beam):
        """
        :param SearchBeam beam:
        :rtype: BatchInfo
        """
        assert self.virtual_dims
        if self.beam == beam:
            return self
        if beam.name in self._descendants_by_beam_name:
            return self._descendants_by_beam_name[beam.name]
        return BatchInfo(
            base=self,
            new_dim=self._make_beam_dim(beam),
            new_dim_index=self.virtual_dims.index(self.get_global_batch_dim()) + 1,
        )

    def copy_remove_beam(self):
        """
        :rtype: BatchInfo
        """
        if not self.beam:
            return self
        assert self.virtual_dims
        root = self.get_global_base()
        dims_wo_beam = [dim for dim in self.virtual_dims if not isinstance(dim, BatchInfo.BeamDim)]
        return root._global_descendants_by_virtual_dims[tuple(dims_wo_beam)]  # must exist

    def copy_remove_dim(self, remove_dim):
        """
        :param VirtualDimBase remove_dim:
        :rtype: BatchInfo
        """
        assert self.virtual_dims
        root = self.get_global_base()
        dims_wo_dim = [dim for dim in self.virtual_dims if dim != remove_dim]
        return root._global_descendants_by_virtual_dims[tuple(dims_wo_dim)]  # must exist

    def copy_set_beam(self, beam):
        """
        :param SearchBeam|None beam:
        :rtype: BatchInfo
        """
        batch = self.copy_remove_beam()
        if beam:
            batch = batch.copy_extend_with_beam(beam)
        return batch

    def copy_extend_with_packed_dim_tag(self, dim_tag, batch_major):
        """
        :param Dim dim_tag:
        :param bool batch_major: if True, add new dim in front. otherwise, add new dim at the end
        :rtype: BatchInfo
        """
        new_dim = self._make_packed_dim(dim_tag)
        new_dim_idx = -1 if batch_major else self._next_spatial_major_index()
        return self._copy_extend_dim(new_dim=new_dim, new_dim_idx=new_dim_idx)

    def copy_extend_with_padded_dim_tag(self, dim_tag, batch_major=None, new_dim_idx=None):
        """
        :param Dim dim_tag:
        :param bool|None batch_major: if True, add new dim in front. otherwise, add new dim at the end
        :param int|None new_dim_idx:
        :rtype: BatchInfo
        """
        new_dim = self._make_padded_dim(dim_tag)
        if new_dim_idx is None:
            assert batch_major is not None
            new_dim_idx = -1 if batch_major else self._next_spatial_major_index()
        else:
            assert batch_major is None
        return self._copy_extend_dim(new_dim=new_dim, new_dim_idx=new_dim_idx)

    def copy_extend_with_padded_or_fixed_dim_tag(self, dim_tag, batch_major=None, new_dim_idx=None):
        """
        :param Dim dim_tag:
        :param bool|None batch_major: if True, add new dim in front. otherwise, add new dim at the end
        :param int|None new_dim_idx:
        :rtype: BatchInfo
        """
        if dim_tag.dyn_size is not None:
            new_dim = self._make_padded_dim(dim_tag)
        else:
            new_dim = BatchInfo.FixedDim(size=dim_tag.get_dim_value(), dim_tag=dim_tag)
        if new_dim_idx is None:
            assert batch_major is not None
            new_dim_idx = -1 if batch_major else self._next_spatial_major_index()
        else:
            assert batch_major is None
        return self._copy_extend_dim(new_dim=new_dim, new_dim_idx=new_dim_idx)

    def _copy_extend_dim(self, new_dim, new_dim_idx):
        """
        :param BatchInfo.VirtualDimBase new_dim:
        :param int new_dim_idx:
        :rtype: BatchInfo
        """
        assert self.virtual_dims
        root = self.get_global_base()
        virtual_dims = list(self.virtual_dims)
        if new_dim_idx < 0:
            assert new_dim_idx == -1
            virtual_dims.append(new_dim)
        else:
            virtual_dims.insert(new_dim_idx, new_dim)
        if tuple(virtual_dims) in root._global_descendants_by_virtual_dims:
            return root._global_descendants_by_virtual_dims[tuple(virtual_dims)]
        return BatchInfo(base=self, new_dim=new_dim, new_dim_index=new_dim_idx)


class SearchBeam:
    """
    Represents info about the beam from some beam search (e.g. via :func:`beam_search`),
    e.g. such as the beam size, but also the dependencies.
    This is somewhat parallel to :class:`SearchChoices`, but simpler,
    and independent from the layers/network (:class:`returnn.tf.layers.base.LayerBase`).
    """

    def __init__(self, beam_size, dependency=NotSpecified, name=None, _next_frame=None):
        """
        :param int beam_size:
        :param SearchBeam|NotSpecified|None dependency:
        :param str|None name:
        :param SearchBeam|None _next_frame:
        """
        if isinstance(dependency, SearchBeam):
            assert name and dependency.name and name != dependency.name
        if name and os.path.basename(name).startswith("prev:"):
            assert _next_frame
        self.beam_size = beam_size
        self.dependency = dependency
        self.name = name
        self._next_frame = _next_frame

    def copy_as_prev_frame(self):
        """
        :rtype: SearchBeam
        """
        if self._next_frame:  # already prev frame -> return self. see logic in RecLayer maybe_transform
            return self
        assert self.name
        name = "%s/prev:%s" % (os.path.dirname(self.name), os.path.basename(self.name))
        return SearchBeam(beam_size=self.beam_size, name=name, _next_frame=self)

    def __repr__(self):
        keys = ["name", "beam_size"]
        if self.dependency is not NotSpecified:
            keys.append("dependency")
        return "%s(%s)" % (self.__class__.__name__, ", ".join(["%s=%r" % (key, getattr(self, key)) for key in keys]))

    def __eq__(self, other):
        """
        :param SearchBeam|object|None other:
        :rtype: bool
        """
        if self is other:
            return True
        if self is None or other is None:
            return False
        if not isinstance(self, SearchBeam) or not isinstance(other, SearchBeam):
            return False
        if self.name is None or other.name is None:
            return False  # cannot identify
        return self.name == other.name

    def __ne__(self, other):
        """
        :param SearchBeam|object|None other:
        :rtype: bool
        """
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def _get_dependency_list(self):
        """
        :return: list as far as it is defined
        :rtype: list[SearchBeam]
        """
        ls = [self]
        while isinstance(ls[-1].dependency, SearchBeam):
            ls.append(ls[-1].dependency)
        return ls

    @classmethod
    def get_combined_beam(cls, beam1, beam2=None, *beams):
        """
        Combines beams.
        This will throw an exception if they cannot be combined.
        Note that in beam search (see :class:`SearchChoices`),
        the logic to combine beams from different search choices
        happens in a generic way for all layers automatically
        via :func:`TFNetwork._create_layer_layer_desc`,
        so normally we already have the same beam.
        Unless we are at template construction.

        :param SearchBeam|None beam1:
        :param SearchBeam|None beam2:
        :param SearchBeam|None beams:
        :rtype: SearchBeam|None
        """
        if beams:
            beam12 = cls.get_combined_beam(beam1, beam2)
            return cls.get_combined_beam(beam12, beams[0], *beams[1:])
        if beam2 is None:
            return beam1
        if beam1 is None:
            return beam2
        if beam1 == beam2:
            if beam2.dependency is NotSpecified:
                return beam1
            if beam1.dependency is NotSpecified:
                return beam2
            return beam1
        assert beam1.name and beam2.name
        if beam2._next_frame and not beam1._next_frame:
            return beam1
        if beam1._next_frame and not beam2._next_frame:
            return beam2
        b1 = beam1
        b2 = beam2
        used_next_frame = False
        if b1._next_frame and b2._next_frame:
            b1 = b1._next_frame
            b2 = b2._next_frame
            used_next_frame = True
        l1 = b1._get_dependency_list()
        l2 = b2._get_dependency_list()
        if b2 in l1:
            return beam1
        if b1 in l2:
            return beam2
        if used_next_frame:
            # Example: beam1: prev:out, beam2: prev:t, t->prev:out (l2).
            if beam1 in l2:  # -> beam1 dep on beam2
                return beam1
            if beam2 in l1:
                return beam2
        raise Exception(
            "\n".join(
                [
                    "Cannot combine beams:",
                    "  1: %s (deps: %s, next %s, next deps %s)"
                    % (
                        beam1,
                        beam1._get_dependency_list(),
                        beam1._next_frame,
                        beam1._next_frame._get_dependency_list() if beam1._next_frame else None,
                    ),
                    "  2: %s (deps: %s, next %s, next deps %s)"
                    % (
                        beam2,
                        beam2._get_dependency_list(),
                        beam2._next_frame,
                        beam2._next_frame._get_dependency_list() if beam2._next_frame else None,
                    ),
                ]
            )
        )



"""
Native code as Python extension module for the RETURNN frontend, including tensor methods and ops.
"""

from __future__ import annotations

import os
import hashlib
from glob import glob
import textwrap
from returnn.util.py_ext_mod_compiler import PyExtModCompiler

_module = None
_my_dir = os.path.dirname(os.path.abspath(__file__))


def get_module(*, verbose: bool = False):
    """
    :return: native Python extension module
    """
    global _module
    if _module and not verbose:
        return _module

    # Single source, which includes all cpp files, and also includes the hash of all hpp/cpp files.
    src_code = ""
    for fn in sorted(glob(_my_dir + "/*.hpp")):
        src_code += f"// {os.path.basename(fn)} code hash md5: {_code_hash_md5(fn)}\n"
    for fn in sorted(glob(_my_dir + "/*.cpp")):
        src_code += f"// {os.path.basename(fn)} code hash md5: {_code_hash_md5(fn)}\n"
        src_code += f'#include "{os.path.basename(fn)}"\n'

    if os.environ.get("RETURNN_TEST") == "1":
        src_code = (
            textwrap.dedent(
                """\
                #define DEBUG 1
                #ifdef NDEBUG
                #undef NDEBUG
                #endif

                """
            )
            + src_code
        )
        verbose = True

    compiler = PyExtModCompiler(
        base_name="_returnn_frontend_native",
        code_version=1,
        code=src_code,
        include_paths=(_my_dir,),
        is_cpp=True,
        verbose=verbose,
    )
    module = compiler.load_py_module()
    if not _module:
        _module = module
    return _module


def _code_hash_md5(filename: str) -> str:
    f_code = open(filename).read()
    h = hashlib.md5()
    h.update(f_code.encode("utf8"))
    return h.hexdigest()


_is_set_up = False


def setup():
    """
    Setup the native code.
    """
    global _is_set_up
    if _is_set_up:
        return
    _is_set_up = True  # only try once

    from returnn.tensor import Tensor, Dim

    # noinspection PyProtectedMember
    from returnn.tensor.tensor import _TensorOpOverloadsMixin, _TensorMixin

    # noinspection PyProtectedMember
    from returnn.tensor.dim import _DimMixin

    # First we can use some existing native variants, which do not require our own native code.
    # The raw_tensor getter is replaced here, the raw_tensor setter is replaced below.
    Tensor.raw_tensor = property(Tensor._raw_tensor.__get__, Tensor.raw_tensor.__set__)  # noqa
    _TensorMixin.placeholder = Tensor.raw_tensor
    Tensor.dims = property(Tensor._dims.__get__)  # noqa
    _TensorMixin.dim_tags = Tensor.dims  # noqa
    _DimMixin.dimension = property(Dim.size.__get__)  # noqa

    try:
        mod = get_module()
    except Exception as exc:
        if os.environ.get("RETURNN_TEST") == "1":
            raise
        print("RETURNN frontend _native backend: Error while getting module:")
        print(exc)
        print("This is optional (although very recommended), so we continue without it.")
        return

    Tensor.raw_tensor = property(Tensor._raw_tensor.__get__, mod.tensor_raw_tensor_setter)  # noqa
    _TensorMixin.placeholder = Tensor.raw_tensor
    _TensorMixin._raw_backend = property(mod.get_backend_for_tensor)  # noqa

    # Functions in _TensorOpOverloadsMixin are replaced with native variants.
    for name, cur_func in _TensorOpOverloadsMixin.__dict__.items():  # just all of them
        if not callable(cur_func):
            continue
        assert name.startswith("__") and name.endswith("__")
        native_func = getattr(mod, "_tensor_" + name[2:-2] + "_instancemethod")
        assert callable(native_func)
        setattr(_TensorOpOverloadsMixin, name, native_func)

    # Functions in _TensorMixin are replaced with native variants.
    for rf_name, native_name in {
        "copy": "tensor_copy",
        "copy_template": "tensor_copy_template",
        "get_out_permutation_to_dims": "tensor_get_out_permutation_to_dims",
        "copy_compatible_to_dims": "tensor_copy_compatible_to_dims",
        "copy_compatible_to_dims_raw": "tensor_copy_compatible_to_dims_raw",
    }.items():
        assert hasattr(_TensorMixin, rf_name)
        native_func = getattr(mod, "_" + native_name + "_instancemethod")
        setattr(_TensorMixin, rf_name, native_func)

    import returnn.frontend as rf
    from returnn.frontend import math_ as rf_math

    # Functions in rf and rf.math_ are replaced with native variants.
    for rf_name, native_name in {
        "compare": "tensor_compare",
        "combine": "tensor_combine",
        "equal": "tensor_eq",
        "not_equal": "tensor_ne",
        "less": "tensor_lt",
        "less_equal": "tensor_le",
        "greater": "tensor_gt",
        "greater_equal": "tensor_ge",
        "add": "tensor_add",
        "sub": "tensor_sub",
        "mul": "tensor_mul",
        "true_divide": "tensor_truediv",
        "floor_divide": "tensor_floordiv",
        "neg": "tensor_neg",
        "mod": "tensor_mod",
        "pow": "tensor_pow",
        "logical_and": "tensor_and",
        "logical_or": "tensor_or",
        "logical_not": "tensor_invert",
        "abs": "tensor_abs",
        "ceil": "tensor_ceil",
        "floor": "tensor_floor",
    }.items():
        assert hasattr(rf, rf_name)
        assert hasattr(rf_math, rf_name)
        native_func = getattr(mod, native_name)
        setattr(rf, rf_name, native_func)
        setattr(rf_math, rf_name, native_func)


_is_set_up_torch = False


def setup_torch():
    """
    Like :func:`setup`, but specifically for the PyTorch backend.
    This assumes that we can `import torch`, unlike :func:`setup`.
    """
    global _is_set_up_torch
    if _is_set_up_torch:
        return
    _is_set_up_torch = True  # only try once

    import torch

    # noinspection PyBroadException
    try:
        mod = get_module()
    except Exception:
        if os.environ.get("RETURNN_TEST") == "1":
            raise
        # No message here, we expect that setup() was already called and printed the message.
        return

    from returnn.torch.frontend import TorchBackend

    TorchBackend.executing_eagerly = True.__bool__
    TorchBackend.get_dtype_name_raw = mod.raw_torch_tensor_get_dtype
    TorchBackend.get_ndim_raw = staticmethod(torch.Tensor.dim)
    TorchBackend.expand_dims_raw = staticmethod(torch.unsqueeze)
    TorchBackend.reshape_raw = staticmethod(torch.reshape)



#!/usr/bin/env python3

"""
This will dump some content from an event log, i.e. summaries/events written via `tf.compat.v1.summary.FileWriter`,
which will be done by our TFEngine.

Also see:
https://stackoverflow.com/questions/37304461/tensorflow-importing-data-from-a-tensorboard-tfevent-file

    tensorboard --inspect --event_file events.out.tfevents...

Usage:

    tf_inspect_summary_log.py events.out.tfevents...

"""

from __future__ import annotations

import os
import sys
from argparse import ArgumentParser

my_dir = os.path.dirname(os.path.abspath(__file__))
returnn_dir = os.path.dirname(my_dir)
sys.path.insert(0, returnn_dir)

import returnn.tf.compat as tf_compat  # noqa


def main():
    """
    Main entry.
    """
    argparser = ArgumentParser()
    argparser.add_argument("file", help="e.g. events.out.tfevents...")
    argparser.add_argument("--tag", default="objective/loss", help="default is 'objective/loss'")
    args = argparser.parse_args()

    print("file: %s" % args.file)
    print("tag: %s" % args.tag)
    for e in tf_compat.v1.train.summary_iterator(args.file):
        for v in e.summary.value:
            if v.tag == args.tag:
                print("step %i: %r" % (e.step, v.simple_value))
    print("done")


if __name__ == "__main__":
    main()



"""
Attention
"""


from __future__ import annotations
from typing import Tuple, Union, Optional, Sequence
import weakref
import logging
from returnn.tensor import Tensor, Dim, single_step_dim
import returnn.frontend as rf


__all__ = [
    "dot_attention",
    "SelfAttentionBase",
    "SelfAttention",
    "CausalSelfAttention",
    "CausalSelfAttentionState",
    "RelPosSelfAttention",
    "RelPosCausalSelfAttention",
    "CrossAttention",
    "LearnedRelativePositionalEncoding",
    "relative_positional_encoding",
    "sinusoidal_positional_encoding",
]


def dot_attention(
    query: Tensor,
    keys: Tensor,
    values: Tensor,
    *,
    key_dim: Dim,
    axis: Dim,
    att_dropout: float = 0.0,
    att_dropout_broadcast: Optional[bool] = None,
) -> Tensor:
    """
    Calculates attention over the given axis, for given key dim.
    Any other unrelated axes do not matter here.
    This can be used for multi-head or single head.
    The query can have other dimensions or not.

    :param query: {..., key_dim}. For self-attention, do not use the `axis` as in `keys` and `values`,
        but rather replace it by another new dim via :func:`replace_dim`.
    :param keys: {..., axis, key_dim}
    :param values: {..., axis}
    :param key_dim: dim in keys and query, to be reduced to calculate the attention energies.
    :param axis: in keys and values, to apply attention on. softmax will be over this axis, and then it will be reduced
    :param att_dropout: dropout for attention weights
    :param att_dropout_broadcast: whether to broadcast over all but ``axis``.
        normally not wanted. disabled by default since behavior version 19.
    :return: like values but with axis removed, and maybe any additional axes from query
    """
    query *= key_dim.dimension**-0.5
    energy = rf.matmul(query, keys, reduce=key_dim)
    att_weights = rf.softmax(energy, axis=axis)
    if att_dropout_broadcast is None:
        att_dropout_broadcast = _att_dropout_broadcast_default()
    att_weights = rf.dropout(att_weights, att_dropout, axis=att_dropout_broadcast and axis)
    # Masking not needed because softmax should already have masked,
    # so we have 0.0 att weights for padded frames.
    att = rf.matmul(att_weights, values, reduce=axis, use_mask=False)
    if values.feature_dim in att.dims:
        att.feature_dim = values.feature_dim
    return att


# noinspection PyAbstractClass
class SelfAttentionBase(rf.Module):
    """
    Shared base class for (non-causal) self attention (:class:`SelfAttention`)
    and causal self attention (:class:`CausalSelfAttention`).

    It uses :func:`dot_attention` for multi-headed dot-attention.
    """

    def __init__(
        self,
        in_dim: Dim,
        proj_dim: Optional[Dim],
        *,
        key_dim_total: Dim,
        value_dim_total: Dim,
        num_heads: Union[int, Dim],
        with_bias: bool = True,
        att_dropout: float = 0.1,
        att_dropout_broadcast: Optional[bool] = None,
    ):
        """
        :param in_dim: input dim
        :param proj_dim: if given, will add a final linear projection to this dim.
            otherwise no projection after the attention
        :param key_dim_total: total key dim. should be a multiple of num_heads
        :param value_dim_total: total value dim. should be a multiple of num_heads
        :param num_heads: number of heads
        :param with_bias: whether to add bias to qkv and proj linear projections.
            Was False in original Transformer, but many recent implementations use True by default.
            Also see: https://github.com/rwth-i6/returnn_common/issues/234.
        :param att_dropout: dropout for attention weights
        :param att_dropout_broadcast: whether to broadcast over all but ``axis``.
            normally not wanted. disabled by default since behavior version 19.
        """
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = proj_dim if proj_dim else value_dim_total
        if isinstance(num_heads, int):
            num_heads = Dim(num_heads, name="num_heads")
        self.key_dim_total = key_dim_total
        self.key_dim_per_head = key_dim_total.div_left(num_heads)
        self.value_dim_total = value_dim_total
        self.value_dim_per_head = value_dim_total.div_left(num_heads)
        self.num_heads = num_heads
        self.qkv_dim_total = 2 * key_dim_total + value_dim_total
        self.qkv_dim_per_head = 2 * self.key_dim_per_head + self.value_dim_per_head
        self.qkv = rf.Linear(in_dim, self.qkv_dim_total, with_bias=with_bias)
        # In Fairseq MultiheadAttention, they use:
        #   nn.init.xavier_uniform_(self.k_proj.weight, gain=1 / math.sqrt(2))  (same for q_proj, v_proj),
        # where xavier_uniform_ means:
        #   std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
        #   a = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation
        #   _no_grad_uniform_(tensor, -a, a)
        # Out nn.init.VarianceScaling with mode="fan_avg", distribution="uniform":
        #   scale = scale * 2.0 / float(fan_in + fan_out)
        #   limit = math.sqrt(3.0 * scale)
        #   nn.random(distribution="uniform", minval=-limit, maxval=limit, ...)
        # Our fan_out is 3 times larger than in Fairseq, because we concatenate q,k,v.
        # Assuming fan_in = fan_out, it means a factor 2 in the denominator.
        # So our default (Glorot, which is VarianceScaling with mode="fan_avg", distribution="uniform", scale=1.0)
        # is already the same as Fairseq.
        # The bias init is different, but not sure how important this is.
        if proj_dim:
            self.proj = rf.Linear(value_dim_total, proj_dim, with_bias=with_bias)
        else:
            self.proj = None
        self.att_dropout = att_dropout
        if att_dropout_broadcast is None:
            att_dropout_broadcast = _att_dropout_broadcast_default()
        self.att_dropout_broadcast = att_dropout_broadcast

    def forward_qkv(self, source: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        :return: q,k,v
        """
        qkv = self.qkv(source)
        qkv = rf.split_dims(qkv, axis=self.qkv_dim_total, dims=(self.num_heads, self.qkv_dim_per_head))
        q, k, v = rf.split(
            qkv,
            axis=self.qkv_dim_per_head,
            out_dims=(self.key_dim_per_head, self.key_dim_per_head, self.value_dim_per_head),
        )
        return q, k, v

    def attention(self, q: Tensor, k: Tensor, v: Tensor, *, kv_axis: Dim) -> Tensor:
        """apply attention"""
        att = dot_attention(
            q,
            k,
            v,
            key_dim=self.key_dim_per_head,
            axis=kv_axis,
            att_dropout=self.att_dropout,
            att_dropout_broadcast=self.att_dropout_broadcast,
        )
        output, _ = rf.merge_dims(att, dims=(self.num_heads, self.value_dim_per_head), out_dim=self.value_dim_total)
        if self.proj:
            output = self.proj(output)
        return output


class SelfAttention(SelfAttentionBase):
    """
    Classic self attention on sequence level
    """

    def __call__(self, source: Tensor, *, axis: Dim) -> Tensor:
        """forward"""
        q, k, v = self.forward_qkv(source)
        kv_axis = Dim(None, name=f"{axis.name}-kv")
        k, _ = rf.replace_dim(k, in_dim=axis, out_dim=kv_axis)
        v, _ = rf.replace_dim(v, in_dim=axis, out_dim=kv_axis)
        return self.attention(q, k, v, kv_axis=kv_axis)


class CausalSelfAttention(SelfAttentionBase):
    """
    Classic causal self attention
    """

    def __call__(
        self,
        source: Tensor,
        axis: Dim,
        *,
        state: Optional[CausalSelfAttentionState] = None,
    ) -> Tuple[Tensor, CausalSelfAttentionState]:
        """forward"""
        q, k, v = self.forward_qkv(source)
        k, v, hist_dim, new_state = _causal_self_att_step(k, v, axis=axis, state=state, self=self)
        output = self.attention(q, k, v, kv_axis=hist_dim)
        return output, new_state

    def default_initial_state(self, *, batch_dims: Sequence[Dim]) -> CausalSelfAttentionState:
        """
        For causal attention.
        """
        # Note: This dim tag is wrong. It should match to the expand_dim inside __call__.
        # So the dim tag itself should be part of the layer state, and we need to define the initial value of it here.
        # This is not really supported, in various ways, also including RETURNN.
        # We just keep this code in place to be prepared for that.
        # The reason it works right now is that we do an optimization where we replace zero init state by 0.
        expand_dim = Dim(0, name="self_att_expand_dim_init")
        return CausalSelfAttentionState(
            k_accum=rf.zeros(list(batch_dims) + [expand_dim, self.num_heads, self.key_dim_per_head]),
            v_accum=rf.zeros(list(batch_dims) + [expand_dim, self.num_heads, self.value_dim_per_head]),
            accum_axis=expand_dim,
        )


def _causal_self_att_step(
    k: Tensor,
    v: Tensor,
    *,
    axis: Dim,
    state: Optional[CausalSelfAttentionState],
    self: rf.Module,
) -> Tuple[Tensor, Tensor, Dim, CausalSelfAttentionState]:
    if axis == single_step_dim:
        assert state, f"{self}: need state for single step"
        k, hist_dim = rf.cum_concat_step(k, prev_accum=state.k_accum, axis=state.accum_axis)
        v, _ = rf.cum_concat_step(v, prev_accum=state.v_accum, out_spatial_dim=hist_dim, axis=state.accum_axis)
    else:
        if state and state.accum_axis.dimension != 0:
            raise NotImplementedError(  # need to concat ...
                f"{self}: on sequence over {axis} with initial state {state} not implemented yet"
            )
        # See CumConcatLayer and https://github.com/rwth-i6/returnn/issues/391 for the idea.
        hist_dim = Dim(rf.range_over_dim(axis, device="cpu") + 1, name=f"{axis.description}:kv")
        k, _ = rf.replace_dim(k, in_dim=axis, out_dim=hist_dim)
        v, _ = rf.replace_dim(v, in_dim=axis, out_dim=hist_dim)
    new_state = CausalSelfAttentionState()
    new_state.k_accum = k
    new_state.v_accum = v
    new_state.accum_axis = hist_dim
    return k, v, hist_dim, new_state


class CausalSelfAttentionState(rf.State):
    """
    State for :class:`StepwiseCausalSelfAttention`.
    """

    def __init__(self, *_args, k_accum: Tensor = None, v_accum: Tensor = None, accum_axis: Dim = None):
        """
        :param k_accum: accumulated keys
        :param v_accum: accumulated values
        :param accum_axis:
        """
        super().__init__(*_args)
        if not _args:
            self.k_accum = k_accum
            self.v_accum = v_accum
            self.accum_axis = accum_axis


class RelPosSelfAttention(SelfAttentionBase):
    """
    Self-attention with relative positional encoding.
    This covers both Shawn et al. self-att rel pos 2018 (https://arxiv.org/abs/1803.02155),
    and Dai et al. Transformer-XL style 2019 (https://arxiv.org/abs/1901.02860).

    It uses :func:`relative_positional_encoding` or :class:`LearnedRelativePositionalEncoding`.

    To get Shawn et al. self-att rel pos 2018 / RETURNN SelfAttentionLayer + RelativePositionalEncodingLayer:
    - with_bias = False (at least that was the RETURNN behavior)
    - with_linear_pos = False
    - with_pos_bias = False
    - learnable_pos_emb = True
    - separate_pos_emb_per_head = False (at least that was the RETURNN default)

    To get Dai et al. Transformer-XL style 2019:
    - with_bias = False would be like the paper, however, in most implementations it is True (default)
    - with_linear_pos = True (default)
    - with_pos_bias = True (default)
    - learnable_pos_emb = True (default)
    - separate_pos_emb_per_head = True (default)

    Further details:
    https://github.com/rwth-i6/returnn_common/wiki/Relative-positional-encoding

    Code references, partly adapted from there:
    https://github.com/espnet/espnet/blob/4138010fb66ad27a43e8bee48a4932829a0847ae/espnet/nets/pytorch_backend/transformer/embedding.py#L260
    https://github.com/kimiyoung/transformer-xl/blob/44781ed21dbaec88b280f74d9ae2877f52b492a5/tf/model.py#L4
    """

    def __init__(
        self,
        in_dim: Dim,
        proj_dim: Optional[Dim],
        *,
        key_dim_total: Dim,
        value_dim_total: Dim,
        num_heads: Union[int, Dim],
        with_bias: bool = True,
        with_linear_pos: bool = True,
        with_pos_bias: bool = True,
        learnable_pos_emb: bool = False,
        learnable_pos_emb_clipping: int = 16,
        separate_pos_emb_per_head: bool = True,
        pos_emb_dropout: float = 0.0,
        att_dropout: float = 0.1,
    ):
        super().__init__(
            in_dim=in_dim,
            proj_dim=proj_dim,
            key_dim_total=key_dim_total,
            value_dim_total=value_dim_total,
            num_heads=num_heads,
            with_bias=with_bias,
            att_dropout=att_dropout,
        )
        self.separate_pos_emb_per_head = separate_pos_emb_per_head
        if with_linear_pos:
            self.pos_emb_feat_dim = self.in_dim
        elif separate_pos_emb_per_head:
            self.pos_emb_feat_dim = self.key_dim_total
        else:
            self.pos_emb_feat_dim = self.key_dim_per_head
        # linear transformation for positional encoding
        self.linear_pos = None
        if with_linear_pos:
            self.linear_pos = rf.Linear(
                self.in_dim, self.key_dim_total if separate_pos_emb_per_head else self.key_dim_per_head, with_bias=False
            )
        self.learned_pos_emb = None
        if learnable_pos_emb:
            self.learned_pos_emb = LearnedRelativePositionalEncoding(
                self.pos_emb_feat_dim, clipping=learnable_pos_emb_clipping
            )
        # these two learnable bias are used in matrix c and matrix d
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        self.pos_bias_u = None
        self.pos_bias_v = None
        if with_pos_bias:
            self.pos_bias_u = rf.Parameter((self.num_heads, self.key_dim_per_head))
            self.pos_bias_v = rf.Parameter((self.num_heads, self.key_dim_per_head))
            self.pos_bias_u.initial = rf.init.Glorot()
            self.pos_bias_v.initial = rf.init.Glorot()
        self.pos_emb_dropout = pos_emb_dropout

    def __call__(self, source: Tensor, *, axis: Dim, **_kwargs) -> Tensor:
        """forward"""
        if self.learned_pos_emb is not None:
            pos_emb, pos_emb_spatial_dim = self.learned_pos_emb(query_spatial_dim=axis, key_value_spatial_dim=axis)
        else:
            pos_emb, pos_emb_spatial_dim = relative_positional_encoding(
                query_spatial_dim=axis, key_value_spatial_dim=axis, feat_dim=self.pos_emb_feat_dim
            )
        if self.pos_emb_dropout:
            pos_emb = rf.dropout(pos_emb, self.pos_emb_dropout)
        if self.linear_pos is not None:
            pos_emb = self.linear_pos(pos_emb)
        if self.separate_pos_emb_per_head:
            pos_emb = rf.split_dims(pos_emb, axis=self.key_dim_total, dims=(self.num_heads, self.key_dim_per_head))
        # pos_emb: (head, 2*time1-1, d_k)

        q, k, v = self.forward_qkv(source)
        hist_dim = Dim(None, name=f"{axis.description}:kv")
        k, _ = rf.replace_dim(k, in_dim=axis, out_dim=hist_dim)
        v, _ = rf.replace_dim(v, in_dim=axis, out_dim=hist_dim)
        q_with_bias_u = (q + self.pos_bias_u) if self.pos_bias_u is not None else q  # (batch, head, time1, d_k)
        q_with_bias_v = (q + self.pos_bias_v) if self.pos_bias_v is not None else q  # (batch, head, time1, d_k)

        # compute attention score
        # first compute matrix a and matrix c
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        # (batch, head, time1, time2)
        matrix_ac = rf.matmul(q_with_bias_u, k, reduce=self.key_dim_per_head)

        # compute matrix b and matrix d
        # (batch, head, time1, 2*time1-1)
        matrix_bd = rf.matmul(q_with_bias_v, pos_emb, reduce=self.key_dim_per_head)
        matrix_bd = _rel_pos_enc_shift(matrix_bd, axis, pos_emb_spatial_dim, hist_dim)

        scores = matrix_ac + matrix_bd  # (batch, head, time1, time2)
        scores *= self.key_dim_per_head.dimension**-0.5
        att_weights = rf.softmax(scores, axis=hist_dim)
        att_weights = rf.dropout(att_weights, self.att_dropout, axis=self.att_dropout_broadcast and hist_dim)
        # Masking not needed because softmax should already have masked,
        # so we have 0.0 att weights for padded frames.
        att = rf.matmul(att_weights, v, reduce=hist_dim, use_mask=False)
        output, _ = rf.merge_dims(att, dims=(self.num_heads, self.value_dim_per_head), out_dim=self.value_dim_total)
        if self.proj:
            output = self.proj(output)
        return output

    # provide this func for compat with some existing code
    @staticmethod
    def _rel_shift(x: Tensor, axis: Dim, pos_emb_spatial_dim: Dim, hist_dim: Dim) -> Tensor:
        return _rel_pos_enc_shift(x, axis, pos_emb_spatial_dim, hist_dim)


def _rel_pos_enc_shift(x: Tensor, axis: Dim, pos_emb_spatial_dim: Dim, hist_dim: Dim) -> Tensor:
    """
    :param x: [B,H,T,T*2-1]
    :param axis: T
    :param pos_emb_spatial_dim: T*2-1
    :param hist_dim: T' (equal to T but separate dim)
    :return: [B,H,T,T']
    """
    batch_dims = x.remaining_dims((axis, pos_emb_spatial_dim))
    x_padded, (pos_emb_spatial_dim_,) = rf.pad(
        x, axes=[pos_emb_spatial_dim], padding=[(1, 0)], value=0.0
    )  # [B,H,T,T*2]

    # Reshape + slice trickery. You need to draw the 2D arrays on paper to understand this.
    # Also see similar trickery in :func:`window`.
    x_padded = rf.reshape(x_padded, (axis, pos_emb_spatial_dim_), (pos_emb_spatial_dim_, axis))  # [B,H,T*2,T]
    x_padded, pos_emb_spatial_dim_ = rf.slice(x_padded, axis=pos_emb_spatial_dim_, start=1)  # [B,H,T*2-1,T]
    x_padded = rf.reshape(x_padded, (pos_emb_spatial_dim_, axis), (axis, pos_emb_spatial_dim_))  # [B,H,T,T*2-1]
    x_padded, _ = rf.slice(x_padded, axis=pos_emb_spatial_dim_, size=hist_dim)  # [B,H,T,T']
    x_padded.verify_out_shape(set(batch_dims) | {axis, hist_dim})
    return x_padded


class RelPosCausalSelfAttention(CausalSelfAttention):
    """
    Self-attention with relative positional encoding.
    This covers both Shawn et al. self-att rel pos 2018 (https://arxiv.org/abs/1803.02155),
    and Dai et al. Transformer-XL style 2019 (https://arxiv.org/abs/1901.02860).

    It uses :func:`relative_positional_encoding` or :class:`LearnedRelativePositionalEncoding`.

    Same defaults as :class:`RelPosSelfAttention`, which is mostly Transformer-XL style.

    Further details:
    https://github.com/rwth-i6/returnn_common/wiki/Relative-positional-encoding

    Code references, partly adapted from there:
    https://github.com/espnet/espnet/blob/4138010fb66ad27a43e8bee48a4932829a0847ae/espnet/nets/pytorch_backend/transformer/embedding.py#L260
    https://github.com/kimiyoung/transformer-xl/blob/44781ed21dbaec88b280f74d9ae2877f52b492a5/tf/model.py#L4
    """

    def __init__(
        self,
        in_dim: Dim,
        proj_dim: Optional[Dim],
        *,
        key_dim_total: Dim,
        value_dim_total: Dim,
        num_heads: Union[int, Dim],
        with_bias: bool = True,
        with_linear_pos: bool = True,
        with_pos_bias: bool = True,
        learnable_pos_emb: bool = False,
        learnable_pos_emb_clipping: int = 16,
        separate_pos_emb_per_head: bool = True,
        pos_emb_dropout: float = 0.0,
        att_dropout: float = 0.1,
    ):
        super().__init__(
            in_dim=in_dim,
            proj_dim=proj_dim,
            key_dim_total=key_dim_total,
            value_dim_total=value_dim_total,
            num_heads=num_heads,
            with_bias=with_bias,
            att_dropout=att_dropout,
        )
        self.separate_pos_emb_per_head = separate_pos_emb_per_head
        if with_linear_pos:
            self.pos_emb_feat_dim = self.in_dim
        elif separate_pos_emb_per_head:
            self.pos_emb_feat_dim = self.key_dim_total
        else:
            self.pos_emb_feat_dim = self.key_dim_per_head
        # linear transformation for positional encoding
        self.linear_pos = None
        if with_linear_pos:
            self.linear_pos = rf.Linear(
                self.in_dim, self.key_dim_total if separate_pos_emb_per_head else self.key_dim_per_head, with_bias=False
            )
        self.learned_pos_emb = None
        if learnable_pos_emb:
            self.learned_pos_emb = LearnedRelativePositionalEncoding(
                self.pos_emb_feat_dim, clipping=learnable_pos_emb_clipping
            )
        # these two learnable bias are used in matrix c and matrix d
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        self.pos_bias_u = None
        self.pos_bias_v = None
        if with_pos_bias:
            self.pos_bias_u = rf.Parameter((self.num_heads, self.key_dim_per_head))
            self.pos_bias_v = rf.Parameter((self.num_heads, self.key_dim_per_head))
            self.pos_bias_u.initial = rf.init.Glorot()
            self.pos_bias_v.initial = rf.init.Glorot()
        self.pos_emb_dropout = pos_emb_dropout

    def __call__(
        self, source: Tensor, *, axis: Dim, state: Optional[CausalSelfAttentionState] = None
    ) -> Tuple[Tensor, CausalSelfAttentionState]:
        """forward"""
        q, k, v = self.forward_qkv(source)
        k, v, hist_dim, new_state = _causal_self_att_step(k, v, axis=axis, state=state, self=self)

        if self.learned_pos_emb is not None:
            pos_emb, pos_emb_spatial_dim = self.learned_pos_emb(query_spatial_dim=axis, key_value_spatial_dim=hist_dim)
        else:
            pos_emb, pos_emb_spatial_dim = relative_positional_encoding(
                query_spatial_dim=axis, key_value_spatial_dim=hist_dim, feat_dim=self.pos_emb_feat_dim
            )
        if self.pos_emb_dropout:
            pos_emb = rf.dropout(pos_emb, self.pos_emb_dropout)
        if self.linear_pos is not None:
            pos_emb = self.linear_pos(pos_emb)
        if self.separate_pos_emb_per_head:
            pos_emb = rf.split_dims(pos_emb, axis=self.key_dim_total, dims=(self.num_heads, self.key_dim_per_head))
        # pos_emb: (head, 2*time1-1, d_k)

        q_with_bias_u = (q + self.pos_bias_u) if self.pos_bias_u is not None else q  # (batch, head, time1, d_k)
        q_with_bias_v = (q + self.pos_bias_v) if self.pos_bias_v is not None else q  # (batch, head, time1, d_k)

        # compute attention score
        # first compute matrix a and matrix c
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        # (batch, head, time1, time2)
        matrix_ac = rf.matmul(q_with_bias_u, k, reduce=self.key_dim_per_head)

        # compute matrix b and matrix d
        # (batch, head, time1, 2*time1-1)
        matrix_bd = rf.matmul(q_with_bias_v, pos_emb, reduce=self.key_dim_per_head)
        matrix_bd = _rel_pos_enc_shift(matrix_bd, axis, pos_emb_spatial_dim, hist_dim)

        scores = matrix_ac + matrix_bd  # (batch, head, time1, time2)
        scores *= self.key_dim_per_head.dimension**-0.5
        att_weights = rf.softmax(scores, axis=hist_dim)
        att_weights = rf.dropout(att_weights, self.att_dropout, axis=self.att_dropout_broadcast and hist_dim)
        # Masking not needed because softmax should already have masked,
        # so we have 0.0 att weights for padded frames.
        att = rf.matmul(att_weights, v, reduce=hist_dim, use_mask=False)
        output, _ = rf.merge_dims(att, dims=(self.num_heads, self.value_dim_per_head), out_dim=self.value_dim_total)
        if self.proj:
            output = self.proj(output)
        return output


class CrossAttention(rf.Module):
    """
    Cross attention

    It uses :func:`dot_attention` for multi-headed dot-attention.
    """

    def __init__(
        self,
        encoder_dim: Dim,
        query_in_dim: Dim,
        proj_dim: Optional[Dim],
        *,
        key_dim_total: Dim,
        value_dim_total: Dim,
        num_heads: Union[int, Dim],
        with_bias: bool = True,
        att_dropout: float = 0.1,
        att_dropout_broadcast: Optional[bool] = None,
    ):
        """
        :param encoder_dim: encoder output dim = input dim for key-value
        :param query_in_dim: input dim for query
        :param proj_dim: if given, will add a final linear projection to this dim.
            otherwise no projection after the attention
        :param key_dim_total: total key dim. should be a multiple of num_heads
        :param value_dim_total: total value dim. should be a multiple of num_heads
        :param num_heads: number of heads
        :param with_bias: whether to add bias to qkv and proj linear projections.
            Was False in original Transformer, but many recent implementations use True by default.
            Also see: https://github.com/rwth-i6/returnn_common/issues/234.
        :param att_dropout: dropout for attention weights
        :param att_dropout_broadcast: whether to broadcast over all but ``axis``.
            normally not wanted. disabled by default since behavior version 19.
        """
        super().__init__()
        self.encoder_dim = encoder_dim
        self.query_in_dim = query_in_dim
        self.out_dim = proj_dim if proj_dim else value_dim_total
        if isinstance(num_heads, int):
            num_heads = Dim(num_heads, name="num_heads")
        self.key_dim_total = key_dim_total
        self.key_dim_per_head = key_dim_total.div_left(num_heads)
        self.value_dim_total = value_dim_total
        self.value_dim_per_head = value_dim_total.div_left(num_heads)
        self.num_heads = num_heads
        self.kv_dim_total = key_dim_total + value_dim_total
        self.kv_dim_per_head = self.key_dim_per_head + self.value_dim_per_head
        self.kv = rf.Linear(encoder_dim, self.kv_dim_total, with_bias=with_bias)
        self.q = rf.Linear(query_in_dim, self.key_dim_total, with_bias=with_bias)
        # In Fairseq MultiheadAttention, they use:
        #   nn.init.xavier_uniform_(self.k_proj.weight, gain=1 / math.sqrt(2))  (same for q_proj, v_proj),
        # where (Xavier Glorot) xavier_uniform_ means:
        #   std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
        #   a = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation
        #   _no_grad_uniform_(tensor, -a, a)
        # Out nn.init.VarianceScaling with mode="fan_avg", distribution="uniform":
        #   scale = scale * 2.0 / float(fan_in + fan_out)
        #   limit = math.sqrt(3.0 * scale)
        #   nn.random(distribution="uniform", minval=-limit, maxval=limit, ...)
        # Xavier Glorot: VarianceScaling with mode="fan_avg", distribution="uniform", scale=1.0.
        self.kv.weight.initial = rf.init.Glorot(scale=3 / 4)
        self.q.weight.initial = rf.init.Glorot(scale=1 / 2)
        # The bias init is different, but not sure how important this is.
        if proj_dim:
            self.proj = rf.Linear(value_dim_total, proj_dim, with_bias=with_bias)
        else:
            self.proj = None
        self.att_dropout = att_dropout
        if att_dropout_broadcast is None:
            att_dropout_broadcast = _att_dropout_broadcast_default()
        self.att_dropout_broadcast = att_dropout_broadcast

    def transform_encoder(self, encoder: Tensor, *, axis: Dim) -> rf.State:
        """
        Transformer encoder output. This is intended as an initial API suggestion.
        """
        k, v = self.forward_kv(encoder)
        return rf.State(k=k, v=v, kv_axis=axis)

    def forward_kv(self, source: Tensor) -> Tuple[Tensor, Tensor]:
        """
        This would be calculated once for the whole sequence (batch)
        and then always reused for :func:`attention`.

        :return: k,v
        """
        qkv = self.kv(source)
        qkv = rf.split_dims(qkv, axis=self.kv_dim_total, dims=(self.num_heads, self.kv_dim_per_head))
        k, v = rf.split(
            qkv,
            axis=self.kv_dim_per_head,
            out_dims=(self.key_dim_per_head, self.value_dim_per_head),
        )
        return k, v

    def forward_query(self, source: Tensor) -> Tensor:
        """
        This is calculated for every different query.

        :return: q
        """
        q = self.q(source)
        q = rf.split_dims(q, axis=self.key_dim_total, dims=(self.num_heads, self.key_dim_per_head))
        return q

    def __call__(self, q: Tensor, encoder: rf.State) -> Tensor:
        q = self.forward_query(q)
        return self.attention(q=q, **encoder)

    def attention(self, q: Tensor, k: Tensor, v: Tensor, *, kv_axis: Dim) -> Tensor:
        """apply attention"""
        att = dot_attention(
            q,
            k,
            v,
            key_dim=self.key_dim_per_head,
            axis=kv_axis,
            att_dropout=self.att_dropout,
            att_dropout_broadcast=self.att_dropout_broadcast,
        )
        output, _ = rf.merge_dims(att, dims=(self.num_heads, self.value_dim_per_head), out_dim=self.value_dim_total)
        if self.proj:
            output = self.proj(output)
        return output


class LearnedRelativePositionalEncoding(rf.Module):
    """
    Learnable relative positional encoding.

    E.g. as used in Shawn et al, 2018 (https://arxiv.org/abs/1803.02155).

    https://github.com/rwth-i6/returnn_common/wiki/Relative-positional-encoding
    """

    def __init__(self, feat_dim: Dim, *, clipping: int = 16, dtype: Optional[str] = None, causal: bool = False):
        """
        :param feat_dim: feature dim, for the emb matrix and output
        :param clipping: max distance to consider. emb matrix shape is [2 * clipping + 1, feat_dim] if not causal,
            else [clipping + 1, feat].
            The first and last frame will be the clipping frames.
        :param dtype: for the emb matrix and output
        """
        super(LearnedRelativePositionalEncoding, self).__init__()
        self.feat_dim = feat_dim
        self.clipping = clipping
        self.clipped_spatial_dim = Dim((1 if causal else 2) * clipping + 1, name="learned-rel-pos")
        self.causal = causal
        self.pos_emb = rf.Parameter((self.clipped_spatial_dim, self.feat_dim), dtype=dtype)

    def __call__(
        self,
        *,
        query_spatial_dim: Dim,
        key_value_spatial_dim: Dim,
        query_offset: Optional[Union[int, Tensor]] = None,
    ) -> Tuple[Tensor, Dim]:
        """
        same interface as :func:`relative_positional_encoding`

        :return: tensor of shape [spatial_dim * 2 - 1, feat_dim], and the out spatial dim (spatial_dim * 2 - 1).
            In the center is the rel pos i-j=0. All to the right are for i-j>0, all to the left for i-j<0.
        """
        indices, out_spatial_dim = _make_indices(query_spatial_dim, key_value_spatial_dim, query_offset)
        indices = rf.clip_by_value(indices, -self.clipping, 0 if self.causal else self.clipping)
        # Shift values to be >= 0. Each integer still uniquely identifies a relative position difference.
        indices = indices + self.clipping

        encoding = rf.gather(self.pos_emb, indices=indices, axis=self.clipped_spatial_dim)  # [out_spatial_dim,n_out]
        return encoding, out_spatial_dim

    def full_matrix(
        self,
        *,
        query_spatial_dim: Dim,
        key_value_spatial_dim: Dim,
        query_offset: Optional[Union[int, Tensor]] = None,
    ) -> Tensor:
        """
        :return: as full matrix [query_spatial_dim,key_value_spatial_dim,feat_dim].
            however, note that __call__ is usually to be preferred, as this gives a more efficient format.
        """
        # Very similar logic as in __call__, _make_indices.
        kv_pos_vec = rf.range_over_dim(key_value_spatial_dim)  # [kv_len]
        if query_spatial_dim == single_step_dim:
            assert query_offset is None  # not sure if any custom query offset makes sense?
            # Assume the kv are the accumulated history, and query is cur frame of it,
            # corresponding to the last frame of the kv.
            query_offset = key_value_spatial_dim.get_size_tensor() - 1
            indices = kv_pos_vec - query_offset  # [q_len,kv_len]
        else:
            q_pos_vec = rf.range_over_dim(query_spatial_dim)  # [q_len]
            indices = rf.combine_bc(kv_pos_vec, "-", q_pos_vec + query_offset)  # [q_len,kv_len]
        indices = rf.clip_by_value(indices, -self.clipping, 0 if self.causal else self.clipping)
        indices = indices + self.clipping
        encoding = rf.gather(self.pos_emb, indices=indices, axis=self.clipped_spatial_dim)  # [q_len,kv_len,n_out]
        return encoding


def _make_indices(
    query_spatial_dim: Dim,
    key_value_spatial_dim: Dim,
    query_offset: Optional[Union[int, Tensor]] = None,
) -> Tuple[Tensor, Dim]:
    kv_pos_vec = rf.range_over_dim(key_value_spatial_dim)  # [kv_len]

    # See also RelativePositionalEncodingLayer
    if query_spatial_dim == single_step_dim:
        indices = kv_pos_vec
        out_spatial_dim = key_value_spatial_dim
        assert query_offset is None  # not sure if any custom query offset makes sense?
        # Assume the kv are the accumulated history, and query is cur frame of it,
        # corresponding to the last frame of the kv.
        query_offset = key_value_spatial_dim.get_size_tensor() - 1
    else:
        query_spatial_dim_m1 = query_spatial_dim - 1
        q_pos_vec = rf.range_over_dim(query_spatial_dim_m1)  # [q_len-1]

        # We want to have all distances as in rf.combine_bc(kv_pos_vec, "-", q_pos_vec) with shape [q_len,kv_len].
        # We want to store only non-duplicates.
        # The min value is with kv_pos=0, q_pos=q_len-1: -(q_len-1)
        # The max value is with kv_pos=kv_len-1, q_pos=0: k_len-1
        indices, out_spatial_dim = rf.concat(
            (q_pos_vec - query_spatial_dim_m1.get_dim_value_tensor(), query_spatial_dim_m1),
            (kv_pos_vec, key_value_spatial_dim),
        )
    if query_offset is not None:
        indices = indices - query_offset

    return indices, out_spatial_dim


_relative_positional_encoding_cache = weakref.WeakKeyDictionary()  # run ctx -> (spatial_dim, feat_dim) -> enc


def relative_positional_encoding(
    *,
    query_spatial_dim: Dim,
    key_value_spatial_dim: Dim,
    feat_dim: Dim,
    query_offset: int = 0,
    dtype: Optional[str] = None,
) -> Tuple[Tensor, Dim]:
    """
    Implements relative positional encoding, Transformer-XL style (https://arxiv.org/abs/1901.02860),
    as used for example by :class:`RelPosSelfAttention`.

    Code references, partly adapted from there:
    https://github.com/espnet/espnet/blob/4138010fb66ad27a43e8bee48a4932829a0847ae/espnet/nets/pytorch_backend/transformer/embedding.py#L260
    https://github.com/kimiyoung/transformer-xl/blob/44781ed21dbaec88b280f74d9ae2877f52b492a5/tf/model.py#L4

    Note that this encoding is stored in a cache so that it is only calculated once.
    and then reused.

    Note that we could extend the implementation later to also buffer it
    even across mini-batches, like the ESPnet implementation does,
    e.g. by storing it in an auxiliary variable and increasing its size when needed.
    But this is not done yet, to keep the code simple.

    :return: tensor of shape [spatial_dim * 2 - 1, feat_dim], and the out spatial dim (spatial_dim * 2 - 1).
      In the center is the rel pos i-j=0. All to the right are for i-j>0, all to the left for i-j<0.
    """
    if not dtype:
        dtype = rf.get_default_float_dtype()
    cache = _relative_positional_encoding_cache.setdefault(rf.get_run_ctx(), {})
    cache_key = (query_spatial_dim, key_value_spatial_dim, feat_dim, query_offset, dtype)
    if cache_key in cache:
        return cache[cache_key]
    import math

    with rf.control_flow_ctx(None):
        # See also RelativePositionalEncodingLayer, LearnedRelativePositionalEncoding
        indices, out_spatial_dim = _make_indices(query_spatial_dim, key_value_spatial_dim, query_offset)

        feat2_dim = feat_dim.div_left(2)
        div_term = rf.exp(rf.range_over_dim(feat2_dim, dtype=dtype) * -(2.0 * math.log(1e4) / feat_dim.dimension))
        arg_sin = rf.combine_bc(rf.cast(indices, dtype), "*", div_term)
        arg_cos = arg_sin + math.pi / 2.0
        arg, feat_dim_ = rf.concat((arg_sin, feat2_dim), (arg_cos, feat2_dim))
        arg, feat_dim_ = rf.replace_dim(arg, in_dim=feat_dim_, out_dim=feat_dim)
        emb = rf.sin(arg)
        emb.verify_out_shape(
            {out_spatial_dim, feat_dim} if out_spatial_dim != single_step_dim else {feat_dim},
            allow_missing_implicit_dims=True,
        )
        emb.feature_dim = feat_dim
        cache[cache_key] = emb, out_spatial_dim
        return emb, out_spatial_dim


_positional_encoding_cache = weakref.WeakKeyDictionary()  # run ctx -> (spatial_dim, feat_dim) -> enc


def sinusoidal_positional_encoding(
    *,
    spatial_dim: Dim,
    feat_dim: Dim,
    offset: Optional[Union[int, Tensor]] = None,
    dtype: Optional[str] = None,
    device: Optional[str] = None,
) -> Tensor:
    """
    Implements absolute sinusoidal positional encoding.

    Code adopted from :func:`relative_positional_encoding`
    and our TF util :func:`get_positional_encoding`.

    Note that this encoding is stored in a cache so that it is only calculated once.
    and then reused.

    Note that we could extend the implementation later to also buffer it
    even across mini-batches, like the ESPnet implementation does,
    e.g. by storing it in an auxiliary variable and increasing its size when needed.
    But this is not done yet, to keep the code simple.

    :return: tensor of shape [spatial_dim, feat_dim] if spatial_dim != single_step_dim else [feat_dim]
    """
    if not dtype:
        dtype = rf.get_default_float_dtype()
    if not device:
        device = rf.get_default_device()
    cache = _positional_encoding_cache.setdefault(rf.get_run_ctx(), {})
    cache_key = (spatial_dim, feat_dim, offset, dtype, device)
    if cache_key in cache:
        return cache[cache_key]
    import math

    with rf.control_flow_ctx(None):
        # See also RelativePositionalEncodingLayer, LearnedRelativePositionalEncoding
        if spatial_dim == single_step_dim:
            assert offset is not None
            indices = rf.convert_to_tensor(offset, device=device)  # scalar
        else:
            indices = rf.range_over_dim(spatial_dim, device=device)  # [len]
            if offset is not None:
                indices = indices + offset
        indices = rf.copy_to_device(indices, device)

        feat2_dim = feat_dim.div_left(2)
        div_term = rf.exp(
            rf.range_over_dim(feat2_dim, dtype=dtype, device=device) * -(math.log(1e4) / (feat2_dim.dimension - 1))
        )
        arg_sin = rf.combine_bc(rf.cast(indices, dtype), "*", div_term)
        arg_cos = arg_sin + math.pi / 2.0
        arg, feat_dim_ = rf.concat((arg_sin, feat2_dim), (arg_cos, feat2_dim))
        arg, feat_dim_ = rf.replace_dim(arg, in_dim=feat_dim_, out_dim=feat_dim)
        emb = rf.sin(arg)
        emb.verify_out_shape(
            {spatial_dim, feat_dim} if spatial_dim != single_step_dim else {feat_dim}, allow_missing_implicit_dims=True
        )
        emb.feature_dim = feat_dim
        cache[cache_key] = emb
        return emb


_att_dropout_broadcast_shown_warning = False


def _att_dropout_broadcast_default() -> bool:
    from returnn.config import get_global_config
    from returnn.util.basic import BehaviorVersion

    config = get_global_config(raise_exception=False)
    if config:
        opt = config.bool("rf_att_dropout_broadcast", None)
        if opt is not None:
            return opt
        opts = config.bool("rf_dropout_broadcast", None)  # also see :func:`dropout_broadcast_default`
        if opts is not None:
            return opts

    if BehaviorVersion.get() <= 18:
        global _att_dropout_broadcast_shown_warning
        if not _att_dropout_broadcast_shown_warning:
            _att_dropout_broadcast_shown_warning = True
            logging.getLogger("returnn.frontend").warning(
                "Attention dropout uses broadcasting. This is old behavior and likely not what you want."
                " Set config option 'rf_att_dropout_broadcast' to False to disable this,"
                " or switch to a new behavior version >= 19."
                " (This warning is only printed once.)"
            )
        return True  # old default
    return False
